import six

from plivo.utils.validators import *
from plivo.xml import PlivoXMLElement, map_type


class RedirectElement(PlivoXMLElement):
    _name = 'Redirect'
    _nestable = []

    @property
    def method(self):
        return self.__method

    @method.setter
    def method(self, value):
        self.__method = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_method(self, value):
        self.method = value
        return self

    def __init__(
            self,
            content,
            method=None,
    ):
        super(RedirectElement, self).__init__()

        self.content = content
        self.method = method

    def to_dict(self):
        d = {
            'method': self.method,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



# -*- coding: utf-8 -*-
"""
Phlo class
"""
from plivo.base import PlivoResource, PlivoResourceInterface
from plivo.resources.nodes import Node, MultiPartyCall, Member
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Phlo(PlivoResource):
    _name = 'Phlo'
    _identifier_string = 'phlo_id'

    def node(self, node_type, node_id):
        self.node_id = node_id
        return self.client.request(
            'GET', ('phlo', self.phlo_id, node_type, node_id), response_type=Node)

    def multi_party_call(self, node_id):
        self.node_id = node_id
        self.node_type = 'multi_party_call'
        return self.client.request(
            'GET', ('phlo', self.phlo_id, self.node_type, node_id), response_type=MultiPartyCall)

    def run(self, **kwargs):
        return self.client.request('POST', ('account', self.client.session.auth[0], 'phlo', self.phlo_id),
                                   to_param_dict(self.run, kwargs, func_args_check=False))

    @validate_args(
        phlo_id=[of_type(six.text_type)],
        node_id=[of_type(six.text_type)],
        member_id=[of_type(six.text_type)],
    )
    def member(self, phlo_id, node_id,
               member_id, action,
               node_type='conference_bridge'):
        """
        :param phlo_id:
        :param node_id:
        :param member_id:
        :param action:
        :param node_type: default value `conference_bridge`
        :return:
        """
        return self.client.phlo.member(phlo_id, node_id,
                                       member_id, action,
                                       node_type
                                       )


class Phlos(PlivoResourceInterface):
    _resource_type = Phlo

    def __init__(self, client):
        super(Phlos, self).__init__(client)

    @validate_args(phlo_id=[of_type(six.text_type)])
    def get(self, phlo_id):
        self.phlo_id = phlo_id
        return self.client.request(
            'GET', ('phlo', phlo_id), response_type=Phlo)

    @validate_args(
        phlo_id=[of_type(six.text_type)],
        node_id=[of_type(six.text_type)],
        member_id=[of_type(six.text_type)],
    )
    def member(self, phlo_id, node_id,
               member_id, action,
               node_type='conference_bridge'):
        """
        :param phlo_id:
        :param node_id:
        :param member_id:
        :param action:
        :param node_type: default value `conference_bridge`
        :return:
        """
        data = {
            'member_id': member_id,
            'phlo_id': phlo_id,
            'node_id': node_id,
            'node_type': node_type
        }
        member = Member(self.client, data)
        return getattr(member, action)()

    @validate_args(
        phlo_id=[of_type(six.text_type)],
        node_id=[of_type(six.text_type)],
        action=[of_type(six.text_type)],
        member_id=[optional(of_type(six.text_type))]
    )
    def multi_party_call(self, phlo_id, node_id, action,
                        member_id=None, trigger_source=None,
                        to=None, role=None):
        data = {
            'phlo_id': phlo_id,
            'node_id': node_id,
            'node_type': 'multi_party_call',
        }

        multi_party_call = MultiPartyCall(self.client, data)

        if action == 'member':
            if not member_id:
                raise ValidationError(
                    'member_id parameter is required'
                )
            return getattr(multi_party_call, action)(member_id)

        if not trigger_source:
            raise ValidationError(
                'trigger_source parameter is required'
            )

        if not to:
            raise ValidationError(
                'to parameter is required'
            )

        if action == 'call' and not role:
            raise ValidationError(
                'role parameter is required'
            )

        return getattr(multi_party_call, action)(trigger_source, to, role)



# -*- coding: utf-8 -*-
"""
Conference class - along with its list class
"""

from plivo.base import PlivoResource, PlivoResourceInterface
from plivo.exceptions import InvalidRequestError
from plivo.utils import to_param_dict


class ConferenceMember(PlivoResource):
    _name = ''
    _identifier_string = 'member_id'


class Conference(PlivoResource):
    _name = 'Conference'
    _identifier_string = 'conference_name'

    def hangup(self):
        return self.client.conferences.hangup(self.conference_name)

    def member_hangup(self, member_id):
        return self.client.conferences.member_hangup(self.conference_name,
                                                     member_id)

    def member_kick(self, member_id):
        return self.client.conferences.member_kick(self.conference_name,
                                                   member_id)

    def member_mute(self, member_id):
        return self.client.conferences.member_mute(self.conference_name,
                                                   member_id)

    def member_mute_stop(self, member_id):
        return self.client.conferences.member_mute_stop(
            self.conference_name, member_id)

    def member_deaf(self, member_id):
        return self.client.conferences.member_deaf(self.conference_name,
                                                   member_id)

    def member_deaf_stop(self, member_id):
        return self.client.conferences.member_deaf_stop(
            self.conference_name, member_id)

    def member_play(self, member_id, url):
        return self.client.conferences.member_play(self.conference_name,
                                                   member_id, url)

    def member_play_stop(self, member_id):
        return self.client.conferences.member_play_stop(
            self.conference_name, member_id)

    def member_speak(self, member_id, text, voice=None, language=None):
        return self.client.conferences.member_speak(self.conference_name,
                                                    member_id, text,
                                                    **to_param_dict(
                                                        self.member_speak,
                                                        locals()))

    def member_speak_stop(self, member_id):
        return self.client.conferences.member_speak_stop(
            self.conference_name, member_id)

    def record(self,
               file_format=None,
               transcription_type=None,
               transcription_url=None,
               transcription_method=None,
               callback_url=None,
               callback_method=None):
        return self.client.conferences.record(self.conference_name,
                                              **to_param_dict(
                                                  self.member_speak, locals()))

    def record_stop(self):
        return self.client.conferences.record_stop(self.conference_name)


class Conferences(PlivoResourceInterface):
    _resource_type = Conference
    _iterable = False

    def list(self):
        return self.client.request('GET', ('Conference', ))

    def get(self, conference_name):
        return self.client.request('GET', ('Conference', conference_name))

    def delete(self, conference_name):
        return self.client.request('DELETE', ('Conference', conference_name))

    def delete_all(self):
        return self.client.request('DELETE', ('Conference', ))

    def hangup_all(self):
        return self.delete_all()

    def hangup(self, conference_name):
        return self.delete(conference_name)

    def member_speak(self,
                     conference_name,
                     member_id,
                     text,
                     voice=None,
                     language=None):
        return self.client.request(
            'POST',
            ('Conference', conference_name, 'Member', member_id, 'Speak'),
            to_param_dict(self.member_speak, locals()))

    def member_play(self, conference_name, member_id, url):
        return self.client.request(
            'POST',
            ('Conference', conference_name, 'Member', member_id, 'Play'),
            to_param_dict(self.member_play, locals()))

    def member_deaf(self, conference_name, member_id):
        return self.client.request(
            'POST',
            ('Conference', conference_name, 'Member', member_id, 'Deaf'))

    def member_mute(self, conference_name, member_id):
        return self.client.request(
            'POST',
            ('Conference', conference_name, 'Member', member_id, 'Mute'))

    def member_speak_stop(self, conference_name, member_id):
        return self.client.request(
            'DELETE',
            ('Conference', conference_name, 'Member', member_id, 'Speak'))

    def member_play_stop(self, conference_name, member_id):
        return self.client.request(
            'DELETE',
            ('Conference', conference_name, 'Member', member_id, 'Play'))

    def member_deaf_stop(self, conference_name, member_id):
        return self.client.request(
            'DELETE',
            ('Conference', conference_name, 'Member', member_id, 'Deaf'))

    def member_mute_stop(self, conference_name, member_id):
        return self.client.request(
            'DELETE',
            ('Conference', conference_name, 'Member', member_id, 'Mute'))

    def member_kick(self, conference_name, member_id):
        return self.client.request(
            'POST',
            ('Conference', conference_name, 'Member', member_id, 'Kick'))

    def member_hangup(self, conference_name, member_id):
        return self.client.request(
            'DELETE', ('Conference', conference_name, 'Member', member_id))

    def record(self,
               conference_name,
               file_format=None,
               transcription_type=None,
               transcription_url=None,
               transcription_method=None,
               callback_url=None,
               callback_method=None):
        return self.client.request('POST',
                                   ('Conference', conference_name, 'Record'))

    def record_stop(self, conference_name):
        return self.client.request('DELETE',
                                   ('Conference', conference_name, 'Record'))




from plivo.utils.validators import *

from ..base import ListResponseObject, PlivoResource, PlivoResourceInterface, ResponseObject
from ..exceptions import *
from ..utils import *

class NumberPool(PlivoResource):
    _name = 'NumberPool'
    _identifier_string = 'number_pool_id'
    shortcodes = None
    numbers = None
    tollfree = None
    def __init__(self, client, number_pool_id):
        self.shortcodes = Shortcode(client, number_pool_id)
        self.numbers = Numbers(client, number_pool_id)
        self.tollfree = Tollfree(client, number_pool_id)
    
    
class Numbers(PlivoResource):
    _name = 'Numbers'
    _identifier_string = 'number_pool_id'

    def buy_add_number(self,
        pattern=None,
        country_iso2=None,
        type=None, 
        region=None, 
        number=''):
        params = {}
        params['rent'] = 'true'
        if self.number_pool_id != "":   
            if number !="":
                return self.client.request(
                'POST', ('NumberPool',self.number_pool_id,'Number', number),params,
                response_type=None,
                objects_type=None)
            else :
                phonenumberparam = {}
                if pattern:
                    phonenumberparam['pattern'] = pattern
                if type:
                    phonenumberparam['type'] = type
                if country_iso2:
                    phonenumberparam['country_iso'] = country_iso2
                if region:
                    phonenumberparam['region'] = region
                number_response = self.client.request(
                        'GET',
                        ('PhoneNumber', ),
                        phonenumberparam,
                        objects_type=None,
                        response_type=None, )
                if len(number_response)< 1:
                    raise  ResourceNotFoundError('Resource not found')
                number = number_response[0]['number']
                return self.client.request(
                'POST', ('NumberPool',self.number_pool_id,'Number', number),params,
                response_type=None,
                objects_type=None)
        else:
            raise ResourceNotFoundError('Resource not found')
    
    @validate_args(
    number=[of_type(six.text_type)]
    )
    def remove(self,  number, unrent=False):
        params = {}
        params['unrent'] = unrent
        return self.client.request(
            'DELETE', ('NumberPool',self.number_pool_id,'Number', number), params,
            response_type=None,
            objects_type=None)

    @validate_args(
    number=[of_type(six.text_type)]
    )
    def add(self,  number):
        return self.client.request(
            'POST', ('NumberPool',self.number_pool_id,'Number', number),
            response_type=None,
            objects_type=None)
    
    @validate_args(
    number=[of_type(six.text_type)]
    )
    def find(self, number):
        return self.client.request(
            'GET', ('NumberPool',self.number_pool_id,'Number', number),
            response_type=None,
            objects_type=None)
    
    def count(self,
        starts_with=None, 
        country_iso2=None,
        type=None):
        params ={}
        if starts_with:
            params['starts_with'] = starts_with
        if country_iso2:
            params['country_iso2'] = country_iso2
        if type:
            params['type'] = type
        
        try:
            response = self.client.request(
            'GET', ('NumberPool',self.number_pool_id,'Number'),
            params ,response_type=None,
            objects_type=None)
            return response['meta']['total_count']
        except:
            return 0

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
        

    def list(self,
        starts_with=None, 
        country_iso2=None,
        type=None, limit=None, offset=None):
        params ={}
        if starts_with:
            params['starts_with'] = starts_with
        if country_iso2:
            params['country_iso2'] = country_iso2
        if type:
            params['type'] = type
        if limit:
            params['limit']= limit
        if offset:
            params['offset'] = offset
        return self.client.request('GET', ('NumberPool',self.number_pool_id,'Number'),
                params ,response_type=None,
                objects_type=None)
         
class Shortcode(PlivoResource):
    _name = 'Shortcodes'
    _identifier_string = 'number_pool_id'

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self, limit=None, offset=None):
        params = {}
        if limit:
            params['limit']= limit
        if offset:
            params['offset'] = offset
        return self.client.request('GET', ('NumberPool',self.number_pool_id,'Shortcode'),params,
            response_type=None,
            objects_type=None)

    @validate_args(shortcode=[of_type(six.text_type)])
    def find(self, shortcode):
        return self.client.request('GET', ('NumberPool',self.number_pool_id,'Shortcode', shortcode),
            response_type=None,
            objects_type=None)

    @validate_args(
    shortcode=[of_type(six.text_type)]
    )
    def remove(self, shortcode):
        return self.client.request(
            'DELETE', ('NumberPool',self.number_pool_id,'Shortcode', shortcode),
            response_type=None,
            objects_type=None)

class Tollfree(PlivoResource):
    _name = 'Tollfree'
    _identifier_string = 'number_pool_id'


    @validate_args(
    number=[of_type(six.text_type)]
    )
    def add(self,  number):
        return self.client.request(
            'POST', ('NumberPool',self.number_pool_id,'Tollfree', number),
            response_type=None,
            objects_type=None)

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self, limit=None, offset=None):
        params = {}
        if limit:
            params['limit']= limit
        if offset:
            params['offset'] = offset
        return self.client.request('GET', ('NumberPool',self.number_pool_id,'Tollfree'),params,
            response_type=None,
            objects_type=None)

    @validate_args(tollfree=[of_type(six.text_type)])
    def find(self, tollfree):
        return self.client.request('GET', ('NumberPool',self.number_pool_id,'Tollfree', tollfree),
            response_type=None,
            objects_type=None)

    @validate_args(
    tollfree=[of_type(six.text_type)]
    )
    def remove(self, tollfree, unrent=False):
        params = {}
        params['unrent'] = unrent
        return self.client.request(
            'DELETE', ('NumberPool',self.number_pool_id,'Tollfree', tollfree), params,
            response_type=None,
            objects_type=None)




# -*- coding: utf-8 -*-
"""
Account & Subaccount classes - along with their list classes
"""

from plivo.base import (ListResponseObject,
                        PlivoResource,
                        PlivoResourceInterface)
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Subaccount(PlivoResource):
    _name = 'Subaccount'
    _identifier_string = 'auth_id'

    def update(self, name, enabled=False):
        return self.client.subaccounts.update(self.id, name, enabled)

    def delete(self, cascade=None):
        return self.client.subaccounts.delete(self.id, cascade)


class Subaccounts(PlivoResourceInterface):
    _resource_type = Subaccount

    @validate_args(auth_id=[is_subaccount_id()])
    def get(self, auth_id):
        return self.client.request(
            'GET', ('Subaccount', auth_id), response_type=Subaccount)

    @validate_args(
        name=[of_type(six.text_type)], enabled=[of_type_exact(bool)])
    def create(self, name, enabled=False):
        return self.client.request(
            'POST', ('Subaccount', ),
            to_param_dict(self.create, locals()),
            response_type=Subaccount)

    @validate_args(
        auth_id=[is_subaccount_id()],
        name=[of_type(six.text_type)],
        enabled=[optional(of_type_exact(bool))])
    def update(self, auth_id, name, enabled=False):
        return self.client.request('POST', ('Subaccount', auth_id),
                                   to_param_dict(self.update, locals()))

    @validate_args(
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ])
    def list(self, limit=20, offset=0):
        return self.client.request(
            'GET', ('Subaccount', ),
            to_param_dict(self.list, locals()),
            response_type=ListResponseObject,
            objects_type=Subaccount)

    @validate_args(
        auth_id=[is_subaccount_id()],
        cascade=[
            optional(of_type_exact(bool))
        ])
    def delete(self, auth_id, cascade=None):
        return self.client.request(
            'DELETE', ('Subaccount', auth_id),
            to_param_dict(self.delete, locals())
        )


class Account(PlivoResource):
    _name = 'Account'
    _identifier_string = 'auth_id'

    def get(self):
        return self.client.account.get()

    def update(self, name=None, city=None, address=None):
        id = self.id
        self.__dict__.update(to_param_dict(self.update, locals()))
        return self.client.account.update(*to_param_dict(
            self.update, locals()))


class Accounts(PlivoResourceInterface):
    _resource_type = Account

    def get(self):
        return self.client.request('GET', tuple(), response_type=Account)

    @validate_args(
        name=[optional(of_type(six.text_type))],
        city=[optional(of_type(six.text_type))],
        address=[optional(of_type(six.text_type))])
    def update(self, name=None, city=None, address=None):
        if not (name or city or address):
            raise ValidationError(
                'One parameter of name, city and address is required')
        return self.client.request(
            'POST', tuple(), {'name': name,
                              'city': city,
                              'address': address})



# -*- coding: utf-8 -*-
"""
Base classes, used to deal with every Plivo resource.
"""

import pprint

from plivo.exceptions import InvalidRequestError


class Meta:
    def __init__(self):
        self.limit = None
        self.next = None
        self.offset = None
        self.previous = None
        self.total_count = None


class PlivoGenericResponse(object):
    """A generic response to cover it all!

    This provides a generic blanket response based on what is received from
    Plivo servers.

    This will be used only during POST and DELETE requests.
    """

    def __init__(self, params_dict, id_string=None):
        for i in params_dict:
            self.__dict__[i] = params_dict[i]

        if id_string and id_string in params_dict:
            self.__dict__['id'] = params_dict[id_string]


class ResponseObject(object):
    def __init__(self, dct):
        self.__dict__.update(dct)

    def __contains__(self, item):
        return item in self.__dict__

    def __getitem__(self, item):
        try:
            return self.__dict__.__getitem__(item)
        except KeyError:
            return self.objects.__getitem__(item)

    def __setitem__(self, key, value):
        self.__dict__.__setitem__(key, value)

    def __delitem__(self, key):
        del self.__dict__

    def __str__(self):
        return pprint.pformat(self.__dict__)

    def __repr__(self):
        return pprint.pformat(self.__dict__)


class ListResponseObject(ResponseObject):
    def __init__(self, client, dct):
        super(ListResponseObject, self).__init__(dct)

    def __iter__(self):
        return self.objects.__iter__()

    def __len__(self):
        return len(self.objects)

    def __str__(self):
        return pprint.pformat(self.objects)

    def __repr__(self):
        return str([object for object in self.objects])


class PlivoResource(ResponseObject):
    """The Plivo resource object

    This provides an interface to deal with all Plivo resources and
    sub-resources
    """

    _identifier_string = None

    @property
    def id(self):
        value = self.__dict__.get(self._identifier_string, None)
        if not value:
            raise ValueError('{} must be set'.format(self._identifier_string))
        return value

    def __init__(self, client, data):
        """Sets up the resource URI along with a hack for Account resource"""

        super(PlivoResource, self).__init__(data)
        self._name = self._name or self.__class__.__name__
        self.client = client

    def __str__(self):
        # return '{class_name}({identifier})'.format(
        #    class_name=self._name, identifier=self.id)

        return pprint.pformat(self.__dict__)

    def __repr__(self):
        return self.__str__()

    def update(self, params, path, **kwargs):
        """
        Test Update
        :param params:
        :param path:
        :param random:
        """
        self.client.request('POST', params, path, **kwargs)

    def _update(self, params):
        if not self.id:
            raise InvalidRequestError(
                'Cannot update a {resource_type} resource without an '
                'identifier'.format(resource_type=self._name))

        response_json = self.client.send_request(
            self.__resource_uri, method='POST', data=params)

        for key in params:
            self.__dict__[key] = params[key]

        for key in response_json:
            self.__dict__[key] = response_json[key]

        return self

    def _execute_action(self,
                        action=None,
                        method='GET',
                        params=None,
                        parse=False):
        if not action:
            response = self.client.send_request(
                self.__resource_uri, method=method, data=params)
        else:
            response = self.client.send_request(
                self.__resource_uri + action + '/', method=method, data=params)

        if not parse:
            return response

        self.__resource_json = response
        self.__parse_json()

        try:
            self.id = response[self._identifier_string]
        except AttributeError:
            pass

        if method == 'POST':
            self.__resource_json = params
            self.__parse_json()
        return self

    def delete(self):
        if not self.id:
            raise InvalidRequestError(
                'Cannot delete a {resource_type} resource without an '
                'identifier'.format(resource_type=self._name))

        return PlivoGenericResponse(
            self.client.send_request(self.__resource_uri, method='DELETE'))

    def get(self, params=None):
        if not self.id:
            raise InvalidRequestError(
                'Cannot get a {resource_type} resource without an '
                'identifier'.format(resource_type=self._name))

        self.__resource_json = self.client.send_request(
            self.__resource_uri, data=params)
        self.__parse_json()
        return self

    def create(self, params):
        if self.id:
            raise InvalidRequestError(
                'Cannot create a {resource_type} resource because another'
                ' {resource_type} resource exists with the same '
                'identifier: {identifier}.'.format(
                    resource_type=self._name, identifier=self.id))

        id_string = None
        if self._identifier_string:
            id_string = self._identifier_string

        return PlivoGenericResponse(
            self.client.send_request(
                self.__resource_uri, data=params, method='POST'), id_string)


class PlivoResourceInterface(object):
    _iterable = True

    def __init__(self, client, **kwargs):
        self.client = client

    def __iter__(self):
        if not getattr(self, 'list') or not self.__class__._iterable:
            raise NotImplementedError(
                'list is not supported for this resource')

        def gen():
            limit = 20
            offset = 0
            while True:
                response = self.list(limit=limit, offset=offset)
                if not response.objects:
                    return

                for item in response:
                    yield item
                offset += limit

        return gen()



import sys
from hmac import new as sign
from hashlib import sha256
from .validators import *

PY2 = True if sys.version_info.major == 2 else False

if PY2:
    from urlparse import urlparse, urlunparse, parse_qs
    from base64 import encodestring as encode
else:
    from urllib.parse import urlparse, urlunparse, parse_qs
    from base64 import encodebytes as encode



def string_format(value):
    if isinstance(value, bytes):
        return ''.join(chr(x) for x in bytearray(value))
    if isinstance(value, (int, float, bool)):
        return str(value)
    if isinstance(value, list):
        return [string_format(x) for x in value]
    if PY2 == True and isinstance(value, unicode):
        return value.decode('utf-8')
    return value


def get_map_from_query(query):
    res_map = dict()
    for key, value in parse_qs(query, keep_blank_values=True).items():
        res_map[string_format(key)] = string_format(value)
    return res_map


def get_sorted_query_string(params):
    keys = sorted(params.keys())
    res_params = []
    for key in keys:
        value = params[key]
        if isinstance(value, list):
            res_params.append(
                '&'.join(['{}={}'.format(string_format(key), val) for val in sorted(string_format(value))]))
        else:
            res_params.append('{}={}'.format(string_format(key), string_format(value)))
    return '&'.join(res_params)


def get_sorted_params_string(params):
    keys = sorted(params.keys())
    res_params = []
    for key in keys:
        value = params[key]
        if isinstance(value, list):
            res_params.append(
                ''.join(['{}{}'.format(string_format(key), val) for val in sorted(string_format(value))]))
        elif isinstance(value, dict):
            res_params.append('{}{}'.format(string_format(key), get_sorted_params_string(value)))
        else:
            res_params.append('{}{}'.format(string_format(key), string_format(value)))
    return ''.join(res_params)


def construct_get_url(uri, params, empty_post_params=True):
    parsed_uri = urlparse(uri.encode('utf-8'))
    base_url = urlunparse((parsed_uri.scheme.decode('utf-8'),
                           parsed_uri.netloc.decode('utf-8'),
                           parsed_uri.path.decode('utf-8'), '', '',
                           '')).encode('utf-8')

    params.update(get_map_from_query(parsed_uri.query))
    query_params = get_sorted_query_string(params)
    if len(query_params) > 0 or not empty_post_params:
        base_url = base_url + bytearray('?' + query_params, 'utf-8')
    if len(query_params) > 0 and not empty_post_params:
        base_url = base_url + bytearray('.', 'utf-8')
    return base_url


def construct_post_url(uri, params):
    base_url = construct_get_url(uri, dict(), True if len(params) == 0 else False)
    return base_url + bytearray(get_sorted_params_string(params), 'utf-8')


def get_signature_v3(auth_token, base_url, nonce):
    base_url = bytearray('{}.{}'.format(string_format(base_url), string_format(nonce)), 'utf-8')
    try:
        return encode(sign(auth_token, base_url, sha256).digest()).strip()
    except TypeError:
        return encode(sign(bytearray(auth_token, 'utf-8'), base_url, sha256).digest()).strip()


@validate_args(
    method=[all_of(of_type(six.text_type), is_in(('POST', 'GET'), case_sensitive=False))],
    uri=[is_url()],
    params=[optional(of_type(dict))],
    nonce=[of_type(six.text_type)],
    auth_token=[of_type(six.text_type)],
    v3_signature=[of_type(six.text_type)],
)
def validate_v3_signature(method, uri, nonce, auth_token, v3_signature, params=None):
    """
        Validates V3 Signature received from Plivo to your server

        :param method: Your callback method
        :param uri: Your callback URL
        :param params: Params received in callback from Plivo
        :param nonce: X-Plivo-Signature-V3-Nonce header
        :param v3_signature: X-Plivo-Signature-V3 header
        :param auth_token: (Sub)Account auth token

        :return: True if the request matches signature, False otherwise
    """
    if params is None:
        params = dict()
    auth_token = bytes(auth_token.encode('utf-8'))
    nonce = bytes(nonce.encode('utf-8'))
    v3_signature = bytes(v3_signature.encode('utf-8'))
    base_url = construct_get_url(uri, params) if method == 'GET' else construct_post_url(uri, params)
    signature = get_signature_v3(auth_token, base_url, nonce)
    return signature in v3_signature.split(b',')
